//-------------------------------------------------------------------------------------
//	dummy_12ax7_v04.c
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//	Phil le   26/01/2013  
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//	includes
//-------------------------------------------------------------------------------------
#define classe_e
#include"dummy_12ax7_v04.h"
#undef classe_e


//-------------------------------------------------------------------------------------
//gestion des interruptions
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//reception d'un message en provenance de la liaison RS232
//-------------------------------------------------------------------------------------
ISR(USART_RXC_vect)
{
	unsigned char cmde;	//commande reçue
	float valeur;		//valeur reçue
	uint8_t ret;		//code retour
	
	//on bloque les interruptions
	cli();
	
	//reception du mot de commande
	ret=rs232_rx_valeur();
	cmde=(uint8_t)valeur_flottante;
	
	//la cmde est elle dans le dictionnaire ?
	// si non, on fait pointer sur une commande "nulle" qui ne fait rien
	if (cmde>ncom)
	{
		cmde=0;
	}

	if (cmde<0)
	{
		cmde=0;
	}

	//reception de la valeur
	ret=rs232_rx_valeur();
	valeur=valeur_flottante;
	
	//juste pour chez moi, j'affiche la commande sur un reseau de leds branché sur le port C
	PORTC=cmde;
	attente=0;
	
	//appel de la routine pointée par la commande
	pf[cmde](valeur);
	
	attente=1;
	
	//on autorise les interruptions
	sei();
}
//-------------------------------------------------------------------------------------


//-------------------------------------------------------------------------------------
//routines
//-------------------------------------------------------------------------------------
//courant du tube 12ax7
uint8_t courant(float va,float vg,float *ia)
{
	//on calcule ia d'après un modèle ultra simplifié
	//le courant de grille sera nul
	//on bride le courant de plaque entre 0 et 5mA
	//en cas de dépassement le code retour est 1 pour ia>5mA et 2 pour ia<0
	
	uint8_t res;
	
	float mu;
	float ex;
	float kg1;
	float kp;
	float kvb;

	mu  = 107.282;
	ex  = 0.900255;
	kg1 = 870.263;
	kp  = 875.025;
	kvb = 6112.42;
	
	//adaptation du modele
	vg=vg+0.5;
	
	*ia = va/kp*log(1+exp(kp*(1/mu+vg/sqrt(kvb+va*va)))) * 1.5;
	
	
	
	//limitations du modèle
	res=0;
	
	if (*ia>IA_max)
	{
		*ia=IA_max;
		res=1;
	}
	
	if (*ia<0)
	{
		*ia=0;
		res=2;
	}
	
	return res;
}

//-------------------------------------------------------------------------------------
//init
void init(void)
{
	uint8_t i;
	//initialisation du tableau de pointeur de fonction
	
	
	//juste pour chez moi, j'affiche la commande sur un reseau de leds branché sur le port C
	DDRC=255;

	i=0;

	pf[i] = fonc_void; i++;	//fonction "nulle"
	pf[i] = arret; i++;		//fonction arret
	
	pf[i] = set_VA_max; i++;
	pf[i] = set_VG2_max; i++;
	pf[i] = set_VG1_max; i++;
	pf[i] = set_VFIL_max; i++;
	pf[i] = set_IA_max; i++;
	pf[i] = set_IG2_max; i++;
	pf[i] = set_IG1_max; i++;
	pf[i] = set_PWA_max; i++;
	pf[i] = set_PWG2_max; i++;
	pf[i] = set_PWG1_max; i++;
	pf[i] = set_VG2_min; i++;
	pf[i] = set_VG1_min; i++;
	pf[i] = set_VFIL_min; i++;
	pf[i] = set_VA; i++;
	pf[i] = set_VG2; i++;
	pf[i] = set_VG1; i++;
	pf[i] = set_VFIL; i++;

	pf[i] = mesure;

	attente=1;	
	afficheur=1;
	sens=1;
}
//-------------------------------------------------------------------------------------


//-------------------------------------------------------------------------------------
//programme principal
//-------------------------------------------------------------------------------------
int main(void)
{	
	//init
	init();
	
	//init rs232 en bidirectionnel à 115200 bauds
	rs232_init(16,1);
	
	//boucle infinie
	while(1)
	{
		if (attente==1)
		{
			chenille();
		}
	}
		
	return 0;
}

//-------------------------------------------------------------------------------------
//routines
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//fonction chenille
//-------------------------------------------------------------------------------------
void chenille(void)
{
	if (sens==1)
	{
		if (afficheur<128)
		{
			afficheur=afficheur*2;
		}
		else
		{
			sens=0;
		}
	}
	else
	{
		if (afficheur>1)
		{
			afficheur=afficheur/2;
		}
		else
		{
			sens=1;
		}
	}

	PORTC=afficheur;
	
	delai(20);
}


//-------------------------------------------------------------------------------------
//fonction nulle
//-------------------------------------------------------------------------------------
void fonc_void(float v)
{
	//on retourne juste un code 99 qui indique que la commande passée
	//n'a pas été interprétée
	
	rs232_tx_valeur(99,0);
}

//-------------------------------------------------------------------------------------
//fonctions set_xx_max et set_xx_min
//
//Ces fonctions ne font que memoriser les valeurs limites à ne pas
//depasser pendant la manip
//
//-------------------------------------------------------------------------------------
void set_VA_max(float v)
{
	VA_max=v;
	rs232_tx_valeur(0,0);
}

//-------------------------------------------------------------------------------------
void set_VG2_max(float v)
{
	VG2_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VG1_max(float v)
{
	VG1_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VFIL_max(float v)
{
	VFIL_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_IA_max(float v)
{
	IA_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_IG2_max(float v)
{
	IG2_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_IG1_max(float v)
{
	IG1_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_PWA_max(float v)
{
	PWA_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_PWG2_max(float v)
{
	PWG2_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_PWG1_max(float v)
{
	PWG1_max=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VG2_min(float v)
{
	VG2_min=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VG1_min(float v)
{
	VG1_min=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VFIL_min(float v)
{
	VFIL_min=v;
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
//fonctions set_xx
//
//Ces fonctions fixent les tensions à appliquer avant chaque mesure
//Elles ont donc une action sur le hard
//
//-------------------------------------------------------------------------------------
void set_VA(float v)
{
	uint8_t ret;
	
	VA=v;
	ret=0;

	//limitations
	if (VA>VA_max)
	{
		ret=1;
	}
	
	//...
	//...
	
	
/*
	action sur le booster à definir
	par exemple:
	
	if (ret==0)
	{
		tension_ok=0;
		nb_tentatives=0;
	
		while (tension_ok==0 && nb_tentative<nb_tentative_max)
		{
			tension_ok=1;
			ret=0;
		
			ecart=mesure_can(voie_va)-VA;
			nb_tentative++;
		
			if (ecart>VA+seuil)
			{
				booster_VA_diminue;
				tension_ok=0;
				ret=2;
			}

			if (ecart<VA-seuil)
			{
				booster_VA_augmente;
				tension_ok=0;
				ret=1;
			}
		
		}
	
		if (nb_tentative>=nb_tentative_max)
		{
			ret=3;
		}
	}
*/	

	rs232_tx_valeur(ret,0);
}
//-------------------------------------------------------------------------------------
void set_VG2(float v)
{
	VG2=v;
	
	//action sur le booster
	//idem reglage de VA
	
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VG1(float v)
{
	VG1=v;
	
	//action sur le booster
	//idem reglage de VA
	
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void set_VFIL(float v)
{
	VFIL=v;
	
	//action sur le booster
	//idem reglage de VA
	
	rs232_tx_valeur(0,0);
}
//-------------------------------------------------------------------------------------
void arret(float v)
{
	rs232_tx_valeur(50,0);
	attente=1;	
}
//-------------------------------------------------------------------------------------
//fonction mesure
//-------------------------------------------------------------------------------------
void mesure(float v)
{
/*
	operation de mesure à definir
	par exemple:
	
	enclenche_HT
	
	//mesure unitaire pour voir si on ne dépasse pas les mim max
	va=mesure_va();
	vg1=mesure_vg1();
	vg2=mesure_vg2();
	vfil=mesure_vfil();
	ia=mesure_ia();
	ig1=mesure_ig1();
	ig2=mesure_ig2();
	
	arret_HT
	
	if (va<va_min)
	{
		ret=1;
	}

	if (va>va_max)
	{
		ret=2;
	}
	
	etc... avec vg1 vg2 et vfil
	
	if (ia>ia_max)
	{
		ret=yy;
	}

	etc... avec ig1 ig2
	
	pwa=ia*1000*va;
	
	if (pwa>pwa_max)
	{
		ret=xx;
	}

	etc... avec pwg1 pwg2
	
	
	if (ret==0)
	{
		//pas de dépassement de limites donc on fait les moyennes
		
		//init
		va_moy=0;
		...
		...
		...
		
		enclenche_HT				
		
		for (i=0;i<nmoy;i++)
		{
			va_moy=va_moy+mesure_va();
			...
			...
			...
		}
		
		arret_HT
		
		//calcul moy
		va_moy=va_moy/nmoy;
		...
		...
		...
		
		
	}
	
	//renvoi des valeurs
	rs232_tx_8(ret);
	rs232_tx_32(va_moy);
	...
	...
	...
*/

	//c'est là que je place mon dummy 12AX7
	//encapsulé dans une boucle
	
	uint8_t ret;
	uint8_t res;
	uint8_t i;
	uint8_t nmoy;
	float va_moy;
	float vg1_moy;
	float vg2_moy;
	float vfil_moy;
	float ia_moy;
	float ig1_moy;
	float ig2_moy;
	uint8_t dep_VA;
	uint8_t dep_VG1;
	uint8_t dep_VG2;
	uint8_t dep_IA;
	uint8_t dep_IG1;
	uint8_t dep_IG2;
	uint8_t dep_PWA;
	uint8_t dep_PWG1;
	uint8_t dep_PWG2;
	float mes_courant;
	float mes_puissance;
	
	//recuperation du nombre de moyennes
	nmoy=(uint8_t)v;

	//limitations du nombre de moyennes
	if (nmoy<=0)
	{
		nmoy=1;
	}

	if (nmoy>100)
	{
		nmoy=100;
	}

	//init	
	va_moy=0;
	vg1_moy=0;
	vg2_moy=0;
	vfil_moy=0;
	ia_moy=0;
	ig1_moy=0;
	ig2_moy=0;
	
	
	
	//verification si depassement des valeurs min max
	dep_VA=0;
	dep_VG1=0;
	dep_VG2=0;
	dep_IA=0;
	dep_IG1=0;
	dep_IG2=0;
	dep_PWA=0;
	dep_PWG1=0;
	dep_PWG2=0;

	ret=0;

	if (VA>VA_max)
	{
		dep_VA=2;
		ret=1;
	}

	if (VG2>VG2_max)
	{
		dep_VG2=2;
		ret=1;
	}

	if (VG2<VG2_min)
	{
		dep_VG2=1;
		ret=1;
	}

	if (VG1>VG1_max)
	{
		dep_VG1=2;
		ret=1;
	}

	if (VG1<VG1_min)
	{
		dep_VG1=1;
		ret=1;
	}

	//si ret!=0 ---> depassement donc pas de mesure
	if (ret==0)
	{
		//mesures unitaires des courants pour verifier des depassements

		//mesure ia	et calcul pwa	
		res=courant(VA,VG1,&mes_courant);
		mes_puissance=VA*mes_courant/1000;
		
		if (mes_courant>IA_max)
		{
			dep_IA=1;
			ret=1;
		}

		if (mes_puissance>PWA_max)
		{
			dep_PWA=1;
			ret=1;
		}

		//mesure ig1 et calcul pwg1
		mes_courant=0;	//pour le dummy 12AX7
		mes_puissance=VG1*mes_courant/1000;
		
		if (mes_courant>IG1_max)
		{
			dep_IG1=1;
			ret=1;
		}

		if (mes_puissance>PWG1_max)
		{
			dep_PWG1=1;
			ret=1;
		}

		//mesure ig2 et pwg2
		mes_courant=0;	//pour le dummy 12AX7
		mes_puissance=VG2*mes_courant/1000;
		
		if (mes_courant>IG2_max)
		{
			dep_IG2=1;
			ret=1;
		}

		if (mes_puissance>PWG2_max)
		{
			dep_PWG2=1;
			ret=1;
		}

		//si ret!=0 ---> depassement donc pas de moyenne des mesures
		if (ret==0)
		{
			//mesures
			for (i=0;i<nmoy;i++)
			{
				va_moy=va_moy+VA;		//pour le dummy 12AX7
				vg2_moy=vg2_moy+VG2;	//pour le dummy 12AX7
				vg1_moy=vg1_moy+VG1;	//pour le dummy 12AX7
	
				ret=courant(VA,VG1,&mes_courant);//pour le dummy 12AX7
				ia_moy=ia_moy+mes_courant;
	
				ig2_moy=ig2_moy+0;	//pour le dummy 12AX7
				ig1_moy=ig1_moy+0;	//pour le dummy 12AX7
			}
			
			//moyennes
			va_moy=va_moy/nmoy;
			vg2_moy=vg2_moy/nmoy;
			vg1_moy=vg1_moy/nmoy;
			ia_moy=ia_moy/nmoy;	
			ig2_moy=ig2_moy/nmoy;	
			ig1_moy=ig1_moy/nmoy;	
		}
		else
		{
			//il y a une demande de mesure avec des parametres valides
			//mais qui engendrent des courants ou puissances hors domaine
			//le vecteur de retour indique quelle variable depasse
			//le code retour donne des indication sur les puissance dépassées
			// 2=pwa 4=pwg1 8=pwg2  (code 11 --> pwamax et pwg2max)

			va_moy=dep_VA;
			vg2_moy=dep_VG2;
			vg1_moy=dep_VG1;
			ia_moy=dep_IA;	
			ig2_moy=dep_IG2;	
			ig1_moy=dep_IG1;
			
			ret=dep_PWA*2+dep_PWG1*4+dep_PWG2*8;
		}
	}
	else
	{
		//il y a une demande de mesure avec des parametres qui sortent du domaine valide
		//le code retour vaut 1 et le vecteur de retour indique quelle variable depasse
		va_moy=dep_VA;
		vg1_moy=dep_VG1;
		vg2_moy=dep_VG2;
	}


	rs232_tx_valeur(ret,0);		//code retour
	rs232_tx_valeur(va_moy,3);	//va
	rs232_tx_valeur(vg2_moy,3);	//vg2
	rs232_tx_valeur(vg1_moy,3);	//vg1
	rs232_tx_valeur(ia_moy,3);	//ia
	rs232_tx_valeur(ig2_moy,3);	//ig2
	rs232_tx_valeur(ig1_moy,3);	//ig1
}

//-------------------------------------------------------------------------------------
//initialisation de la transmission
//-------------------------------------------------------------------------------------
void rs232_init(unsigned char baud,unsigned char isr_rx)
{
	cli();
	
	//reglage de la vitesse
 	UBRRH=(unsigned char)(baud>>8);
	UBRRL=(unsigned char)baud;
	
	//autorisation transmission reception
	UCSRB=(1<<RXEN)|(1<<TXEN);
	
	//reglage protocole (8 bits de donnees, 2 bits de stop)
	UCSRC=(1<<URSEL)|(1<<USBS)|(3<<UCSZ0);
	
	//reglage mode asynchrone
 	UCSRC=UCSRC | 64;
 	
 	//reglage double debit
	UCSRA=UCSRA | 2;
	
	//autorisation interruption sur reception
	if (isr_rx!=0)
	{
		UCSRB = UCSRB | 128;
	}
	
	sei();
}
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//reception d'un mot de 8 bits
//-------------------------------------------------------------------------------------
unsigned char rs232_rx_8(void)
{
	//attente de reception
	while (!(UCSRA & (1<<RXC)))
	{
	}
	
	//reception et retour
	return UDR;
}
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//emission d'un mot de 8 bits
//-------------------------------------------------------------------------------------
void rs232_tx_8(unsigned char v)
{
	//attente buffer emission vide
	while (!(UCSRA & (1<<UDRE)))
	{
	}

	//envoi
	UDR=v;
}
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//emission d'une valeur quelconque
//-------------------------------------------------------------------------------------
void rs232_tx_valeur(float v,int nd)
{
	int E;
	int D;
	int S;
	float V;
	float mul;
	int i;
	
	//determination du signe
	if (v<0)
	{
		S=-1;
		V=-v;
	}
	else
	{
		S=1;
		V=v;
	}
	
	//partie entiere
	E=floor(V);
	
	//coef multiplicateur pour extraction des decimales
	mul=1;
	
	for (i=0;i<nd;i++)
	{
		mul=mul*10;
	}
	
	//partie decimale
	D=floor(V*mul-E*mul);
	
	trace_ent(E,S);
	trace_dec(D,nd);
	
	rs232_tx_8(10);
}
//-------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
//trace_ent
//-------------------------------------------------------------------------------------
void trace_ent(int v,int s)
{
	int i;
	int mag;
	int V;
	int div;
	int A;
	

	V=v;
	
	//magnitude de v
	if (v!=0)
	{
		mag=floor(log10(v));
	}
	else
	{
		mag=0;
	}
	
	//coef diviseur
	div=1;
	
	for (i=0;i<mag;i++)
	{
		div=div*10;
	}
	
	
	//signe
	if (s==-1)
	{
		rs232_tx_8('-');

	}
	
	for (i=mag;i>=0;i--)
	{
		A=floor(V/div);
		
		rs232_tx_8(A+48);
		
		V=V-A*div;
		div=div/10;
	}
}

//-------------------------------------------------------------------------------------
//trace_dec
//-------------------------------------------------------------------------------------
void trace_dec(int v,int nd)
{
	int V;
	int i;
	int div;
	int A;
	
	V=v;
	
	if (V!=0)
	{
		rs232_tx_8('.');	
		
		//coef diviseur
		div=1;
	
		for (i=0;i<nd;i++)
		{
			div=div*10;
		}
	
		for (i=nd;i>0;i--)
		{
			div=(int)(div/10);

			A=floor((float)V/(float)div);
		
			rs232_tx_8(A+48);
		
			V=V-A*div;
		}
	}
}

//-------------------------------------------------------------------------------------
//delai
//-------------------------------------------------------------------------------------
void delai(uint16_t d)
{
	uint16_t i;
	
	for (i=0;i<d;i++)
	{
		_delay_ms(1.0);
	}
}

//-------------------------------------------------------------------------------------
//reception d'une valeur quelconque
//-------------------------------------------------------------------------------------
uint8_t rs232_rx_valeur(void)
{
	uint8_t i;
	uint8_t n;
	uint8_t nc;
	uint8_t suite;
	unsigned char c[30];
	uint8_t valid;
	uint8_t cpt;
	int8_t signe;
	int8_t posv;
	float v;
	float mul;

	uint8_t ct;
	
	n=30;
	suite=1;
	i=0;
	
	//reception
	while (suite==1 && i<n)
	{
		ct=rs232_rx_8();
		
		if (ct==10)
		{
			suite=0;
		}

		c[i]=(unsigned char)ct;
		i++;
	}
	
	nc=i-1;
	

	//analyse
	
	//detection du signe
	if (c[0]=='-')
	{
		signe=-1;
		c[0]='0';
	}
	else
	{
		signe=1;
	}

	valid=1;
	cpt=0;
	
	//test si caractere possible 0123456789.
	//et comptage du nombre de virgules
	posv=-1;
	
	for (i=0;i<nc;i++)
	{
		if (c[i]<'0' || c[i]>'9')
		{
			if (c[i]!='.')
			{
				valid=0;
			}
		}
		
		if (c[i]=='.')
		{
			cpt++;
			posv=i;
		}
	}
	
	if (cpt>1)
	{
		valid=0;
	}
	
	if (valid==1)
	{
		//formation de la valeur
		mul=1;
	
		//cas general de la virgule placée n'importe ou
		for (i=0;i<posv-1;i++)
		{
			mul=mul*10;
		}
	
		//cas particulier de la virgule placée en premier
		if (posv==0)
		{
			mul=0.1;
		}
	
		//cas particulier sans virgule
		if (posv==-1)
		{
			mul=1;
	
			for (i=0;i<nc-1;i++)
			{
				mul=mul*10;
			}
		}
	
		//conversion
		v=0;
	
		for (i=0;i<nc;i++)
		{
			if (c[i]!='.')
			{
				v=v+(c[i]-48)*mul;
				mul=mul/10;
			}
		}
		
		//depassement de capacité
		if (v>=32768)
		{
			valeur_flottante=0;
			valid=0;	
		}
		else
		{
			v=v*signe;
			valeur_flottante=v;
		}
	}
	else
	{
		valeur_flottante=0;
	}
	
	return valid;
}







